/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004-2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: util.c,v 1.7 2005/04/24 15:36:21 kasperd Exp $
*/

#include "tstore.h"
#include "test.h"

void errstop(const char *s)
{
  perror(s);
  exit(EXIT_FAILURE);
}

void close_store(struct store *s)
{
  if (!s) return;
  if (s->locked) pwrite(s->fd,"",1,DIRTYFLAG_OFFSET);
  if (s->fd != -1) close(s->fd);
  if (s->name) free(s->name);
  if (s->h32name) free(s->h32name);
}

struct store *open_store(char *name, int flags, enum lock lock)
{
  struct store *s;
  NEW(s);
  s->locked=0;
  if ((s->name=strdup(name))) {
    if ((s->h32name=malloc(strlen(name)+5))) {
      sprintf(s->h32name,"%s.32h",name);
      if ((s->fd=open(name,flags))!=-1) {
	uint64_t store_size,first_file;
	switch (myread(s->fd,s->header_buf,512)) {
	case -1:
	  break;
	default:
	  errno=EINVAL;
	  break;
	case 512:
	  if (memcmp(s->header_buf,"TSTORE6",8)) {
	    errno=EINVAL;
	  } else {
	    store_size=GET_BNUM(s->header_buf+STORESIZE_OFFSET);
	    first_file=GET_BNUM(s->header_buf+NEXTPTR_OFFSET);
	    s->store_size=store_size;

	    if (first_file >= store_size) {
	      fprintf(stderr,"First file pointer outside store\n");
	      errno=EINVAL;
	    } else {
	      if (lock == DONT_LOCK) return s;
	      assert(lock == CHECK_LOCK);
	      if (s->header_buf[DIRTYFLAG_OFFSET]) {
		errno=EBUSY;
	      } else {
		return s;
	      } /* endif locked */
	    } /* endif first_file */
	  } /* endif memcmp */
	} /* switch myread */
      } /* endif open */
    } /* endif malloc */
  } /* endif strdup */
  close_store(s);
  return NULL;
}

int lock_store(struct store *s)
{
  uint8_t buf[512];
  if (s->locked) return 0;
  switch(pread(s->fd,buf,TS_BLOCK_SIZE,0)) {
  case 512:
    if (memcmp(buf,s->header_buf,TS_BLOCK_SIZE)) {
    default:
      errno=EBUSY;
    case -1:
      return -1;
    }
  }
  /* FIXME: Here I can race against another process also
   *        trying to lock the same store. Still using the
   *        dirtyflag reduces the window significantly.
   */
  if (pwrite(s->fd,"\1",1,DIRTYFLAG_OFFSET)!=1) return -1;
  s->locked=1;
  return 0;
}

ssize_t myread(int fd, void *buf, size_t count)
{
  ssize_t r,total=0;
  memset(buf,0,count);
  while (count) {
    r=read(fd,buf,count);
    if (r<1) return total?total:r;
    total+=r;
    buf=(((char*)buf)+r);
    count-=r;
  }
  return total;
}

uint64_t get_nbyteint(uint8_t *p, int n)
{
  uint64_t r=0;
  int i;
  for (i=0;i<n;++i)
    r=(r<<8)+p[i];
  return r;
}

void put_nbyteint(uint8_t *p, int n, uint64_t v)
{
  int i;
  for (i=0;i<n;++i)
    p[i]=v>>(((n-1)-i)*8);
  assert(get_nbyteint(p,n)==v);
}

int filelevels(uint64_t size)
{
  int levels=0;
  uint64_t maxsize=512;
  while(size>maxsize) {
    levels++;
    maxsize*=102;
  }
  return levels;
}

int seqblock(uint8_t *buf)
{
  uint64_t v;
  int i;
  if (buf[510]||buf[511]||(buf[0]>=0xC0)) return 0;
  v=GET_BNUM(buf);
  for (i=1;i<102;++i)
    if (GET_BNUM(buf+5*i)!=v+i) return 0;
  return 1;
}

void get_block(struct store *s, uint64_t n, uint8_t *buf)
{
  if (n) {
    if (n&(3LL<<38)) {
      int i;
      int delta=0;
      switch(n>>38) {
      case 1: delta=1; break;
      case 2: delta=FANOUT; break;
      case 3: delta=FANOUT*FANOUT; break;
      }
      assert(delta);
      n-=(1LL<<38);
      for (i=0;i<102;++i)
	PUT_BNUM(buf+5*i,n+i*delta);
      buf[510]=buf[511]=0;
    } else {
      switch(pread(s->fd,buf,512,512*n)) {
      case -1:
	errstop(s->name);
      default:
	fprintf(stderr,"%s: short pread (block %lld)\n",s->name,n);
	exit(1);
      case 512:
	break;
      }
    }
  } else
    memset(buf,0,512);
}

int confirm()
{
  char answer[9];
  fgets(answer,6,stdin);
  return !strncasecmp(answer,"yes",3);
}

void cancel()
{
  fprintf(stderr,"Canceled\n");
  exit(EXIT_FAILURE);
}
